"""
    isort:skip_file
"""

import os
import sys
import unittest
import urllib.request
import zipfile

# sys.path.append(os.getcwd()[:-6])
# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
#
#
# parent_dir = os.path.abspath(os.path.join(os.getcwd(), os.pardir))

import astropy.io.fits as pyfits
import simmer.darks as darks
import simmer.drivers as drivers
import simmer.flats as flats
import simmer.image as image
import simmer.insts as i
import numpy as np
import pandas as pd
import simmer.sky as sky


def download_folder(folder, path=None):
    """
    Downloads a .zip file from this projects S3 testing bucket, unzips it,
    and deletes the .zip file.

    Inputs:
        folder : (string) name of the folder to be downloaded.
    """

    folder_dict = {
        "sky_test": "hhs5w81dvok5wp8",
        "dark_test": "ao1ug1kvlr5l4y3",
        "flat_test": "r0gntctnfrjh5zd",
        "PHARO_config_driver": "p3wv7l800fdqx5q",
        "image_test": "9qwa6ojfsk1l5pq",
        "PHARO_image_driver": "p9uivslz7hoym5c",
        "PHARO_integration": "v8l50zm7jbrccqj",
        "shane_quickstart": "q6m6ls2x2186u3p",
        "readpharo_test": "l8fi3100v5flufp",
        "config_test": "q0vqvy1ejd6rn14",
    }

    def retrieve_extract(path):
        with zipfile.ZipFile(folder + ".zip", "r") as zip_ref:
            zip_ref.extractall(path)

    url = f"https://www.dropbox.com/s/{folder_dict[folder]}/{folder}.zip?dl=1"
    u = urllib.request.urlopen(url)
    data = u.read()
    u.close()

    with open(f"{folder}.zip", "wb") as f:
        f.write(data)
    if path:
        retrieve_extract(path)
    elif "src" in os.listdir():  # if we're actually running tests
        retrieve_extract("src/simmer/tests/")
    else:  # we're running this in an arbitrary directory
        retrieve_extract("")
    os.remove(folder + ".zip")


def delete_folder(folder):
    if os.listdir(folder):
        for f in os.listdir(folder):
            if folder[-1] == "/":
                path = folder + f
            else:
                path = folder + "/" + f
            if os.path.isfile(path):
                os.remove(path)
            else:  # if it's a directory!
                delete_folder(path)
    os.rmdir(folder)


class DataDownloadException(Exception):
    """Base class for data download exceptions in this module."""

    pass


class TestCreation(unittest.TestCase):

    inst = i.ShARCS()

    def test_create_darks(self):
        print("Testing darks")

        try:
            download_folder("dark_test")
        except:
            raise DataDownloadException(
                "Could not download test data for darks."
            )

        raw_dir, reddir = (
            "src/simmer/tests/dark_test/",
            "src/simmer/tests/dark_test/",
        )
        compare_dark = pyfits.getdata(raw_dir + "compare_dark.fits", 0)
        zero = np.zeros(np.shape(compare_dark))  # only testing flats
        darklist = range(1357, 1360)
        try:
            result = darks.create_darks(raw_dir, reddir, darklist, self.inst)
            delete_folder(raw_dir)
            self.assertCountEqual(
                np.ravel(result - compare_dark), np.ravel(zero)
            )
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_create_flats(self):
        print("Testing flats")

        try:
            download_folder("flat_test")
        except:
            raise DataDownloadException(
                "Could not download test data for flats."
            )

        raw_dir, reddir = (
            "src/simmer/tests/flat_test/",
            "src/simmer/tests/flat_test/",
        )
        compare_flat = pyfits.getdata(raw_dir + "compare_flat.fits", 0)
        zero = np.zeros(np.shape(compare_flat))  # only testing flats
        flatlist = range(1108, 1114)
        try:
            result = flats.create_flats(
                raw_dir, reddir, flatlist, np.nan, self.inst, test=True
            )

            delete_folder(raw_dir)
            self.assertCountEqual(
                np.ravel(result - compare_flat), np.ravel(zero)
            )
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_create_skies(self):
        print("Testing skies")
        try:
            download_folder("sky_test")
        except:
            raise DataDownloadException(
                "Could not download test data for skies."
            )
        raw_dir, reddir = (
            "src/simmer/tests/sky_test/",
            "src/simmer/tests/sky_test/",
        )
        compare_sky = np.loadtxt(raw_dir + "compare_sky.txt")
        s_dir = raw_dir
        skylist = range(1218, 1222)
        try:
            result = sky.create_skies(
                raw_dir, reddir, s_dir, skylist, self.inst
            )
            zero = np.zeros(np.shape(compare_sky))

            delete_folder(raw_dir)
            self.assertCountEqual(
                np.ravel(result - compare_sky), np.ravel(zero)
            )
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_create_imstack(self):
        print("Testing imstack")

        try:
            download_folder("sky_test")
        except:
            raise DataDownloadException(
                "Could not download test data for skies."
            )
        raw_dir, reddir = (
            "src/simmer/tests/sky_test/",
            "src/simmer/tests/sky_test/",
        )
        imlist = range(1218, 1222)
        s_dir = raw_dir
        try:
            result, shifts_all = image.create_imstack(
                raw_dir, reddir, s_dir, imlist, self.inst
            )
            compare_list = [
                "compare_create_imstack_0",
                "compare_create_imstack_1",
                "compare_create_imstack_2",
                "compare_create_imstack_3",
            ]
            compare_imstack = np.array(
                [np.loadtxt(raw_dir + file) for file in compare_list]
            )
            zero = np.zeros(np.shape(compare_imstack))
            delete_folder(raw_dir)
            self.assertCountEqual(
                np.ravel(result - compare_imstack), np.ravel(zero)
            )
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_create_im_default(self):
        print("Testing default image creation")

        try:
            download_folder("sky_test")
        except:
            raise DataDownloadException(
                "Could not download test data for skies."
            )

        raw_dir, reddir = (
            "src/simmer/tests/sky_test/",
            "src/simmer/tests/sky_test/",
        )
        s_dir = raw_dir
        compare_final_im = pyfits.getdata(
            raw_dir + "Ks/compare_final_im_default.fits", 0
        )
        try:
            image.create_im(s_dir, 10, method="default")
            final_im = pyfits.getdata(raw_dir + "Ks/final_im.fits", 0)
            zero = np.zeros(np.shape(final_im))
            val = np.all(
                np.ravel(final_im - compare_final_im) == np.ravel(zero)
            )
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_create_im_saturated(self):
        print("Testing saturated image creation")

        try:
            download_folder("sky_test")
        except:
            raise DataDownloadException(
                "Could not download test data for skies."
            )

        raw_dir, reddir = (
            "src/simmer/tests/sky_test/",
            "src/simmer/tests/sky_test/",
        )
        s_dir = raw_dir
        compare_final_im = pyfits.getdata(
            raw_dir + "Ks/compare_final_im.fits", 0
        )
        try:
            image.create_im(s_dir, 10, method="saturated")
            final_im = pyfits.getdata(raw_dir + "Ks/final_im.fits", 0)
            zero = np.zeros(np.shape(final_im))
            val = np.all(np.allclose(final_im, compare_final_im))
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(val)


class TestDrivers(unittest.TestCase):
    inst = i.ShARCS()

    def test_dark_driver(self):
        print("Testing dark driver")

        try:
            download_folder("dark_test")
        except:
            raise DataDownloadException(
                "Could not download test data for darks."
            )

        raw_dir, reddir = (
            "src/simmer/tests/dark_test/",
            "src/simmer/tests/dark_test/",
        )
        config = pd.read_csv(os.getcwd() + "/src/simmer/tests/test_config.csv")
        try:
            darks.dark_driver(raw_dir, reddir, config, self.inst)
            val = "dark_3sec.fits" in os.listdir(raw_dir)
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_flat_driver(self):
        print("Testing flat driver")

        try:
            download_folder("flat_test")
        except:
            raise DataDownloadException(
                "Could not download test data for flats."
            )
        raw_dir, reddir = (
            "src/simmer/tests/flat_test/",
            "src/simmer/tests/flat_test/",
        )
        config = pd.read_csv(os.getcwd() + "/src/simmer/tests/test_config.csv")
        source = raw_dir + "flat_J.fits"
        dest = raw_dir + "temp.fits"
        os.rename(source, dest)
        try:
            flats.flat_driver(raw_dir, reddir, config, self.inst)
            val = "flat_J.fits" in os.listdir(raw_dir)
            os.remove(raw_dir + "flat_J.fits")
            os.rename(dest, source)
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_sky_driver(self):
        print("Testing sky driver")

        try:
            download_folder("sky_test")
        except:
            raise DataDownloadException(
                "Could not download test data for skies."
            )

        raw_dir, reddir = (
            "src/simmer/tests/sky_test/",
            "src/simmer/tests/sky_test/",
        )
        config = pd.read_csv(os.getcwd() + "/src/simmer/tests/test_config.csv")
        try:
            sky.sky_driver(raw_dir, reddir, config, self.inst)
            val = "sky.fits" in os.listdir(raw_dir + "/K09203794/Ks")
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_image_driver(self):
        print("Testing image driver")

        try:
            download_folder("image_test")
        except:
            raise DataDownloadException(
                "Could not download test data for images."
            )

        raw_dir, reddir = (
            "src/simmer/tests/image_test/",
            "src/simmer/tests/image_test/",
        )
        config = pd.read_csv(os.getcwd() + "/src/simmer/tests/test_config.csv")
        try:
            method = image.image_driver(raw_dir, reddir, config, self.inst)
            remove_files = [
                "sh00.fits",
                "sh01.fits",
                "sh02.fits",
                "sh03.fits",
                "shifts.txt",
            ]
            val = np.all(
                [
                    r in os.listdir(raw_dir + f"K09203794/Ks")
                    for r in remove_files
                ]
            )
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)


class TestIntegration(unittest.TestCase):
    p = i.PHARO()

    def test_PHARO_all_drivers(self):
        print("Testing PHARO integration")  # need better way to get config?

        try:
            download_folder("PHARO_integration")
        except:
            raise DataDownloadException(
                "Could not download test data for PHARO integration."
            )

        raw_dir, reddir = (
            os.getcwd() + "/src/simmer/tests/PHARO_integration/",
            os.getcwd() + "/src/simmer/tests/PHARO_integration/",
        )
        config_file = os.getcwd() + "/src/simmer/tests/test_image_config.csv"
        try:
            drivers.all_driver(self.p, config_file, raw_dir, reddir)
            compare_final_im = pyfits.getdata(
                raw_dir + "compare_final_im.fits", 0
            )
            final_im = pyfits.getdata(
                raw_dir + "HIP49081/Br-gamma/final_im.fits", 0
            )
            zero = np.zeros(np.shape(final_im))
            val = np.all(
                np.allclose(final_im, compare_final_im, equal_nan=True)
            )

            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_PHARO_config_drivers(self):
        print(
            "Testing PHARO config integration"
        )  # need better way to get config?

        try:
            download_folder("PHARO_config_driver")
        except:
            raise DataDownloadException(
                "Could not download test data for PHARO config integration."
            )
        raw_dir, reddir = (
            os.getcwd() + "/src/simmer/tests/PHARO_config_driver/",
            os.getcwd() + "/src/simmer/tests/PHARO_config_driver/",
        )
        config_file = os.getcwd() + "/src/simmer/tests/test_image_config.csv"
        try:
            drivers.config_driver(self.p, config_file, raw_dir, reddir)
            sky = pyfits.getdata(raw_dir + "HIP49081/Br-gamma/sky.fits", 0)
            compare_sky = pyfits.getdata(
                raw_dir + "HIP49081/Br-gamma/compare_sky.fits", 0
            )
            zero = np.zeros(np.shape(compare_sky))
            val = np.all(np.allclose(compare_sky, sky, equal_nan=True))

            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)

    def test_PHARO_image_drivers(self):
        print(
            "Testing PHARO image integration"
        )  # need to actually point to correct folder, change aodirs

        try:
            download_folder("PHARO_image_driver")
        except:
            raise DataDownloadException(
                "Could not download test data for PHARO image driver."
            )
        raw_dir, reddir = (
            os.getcwd() + "/src/simmer/tests/PHARO_image_driver/",
            os.getcwd() + "/src/simmer/tests/PHARO_image_driver/",
        )
        config_file = os.getcwd() + "/src/simmer/tests/test_image_config.csv"
        try:
            drivers.image_driver(self.p, config_file, raw_dir, reddir)
            remove_files = [
                "sh00.fits",
                "sh01.fits",
                "sh02.fits",
                "sh03.fits",
                "shifts.txt",
            ]
            val = np.all(
                [
                    r in os.listdir(raw_dir + "HIP49081/Br-gamma/")
                    for r in remove_files
                ]
            )

            file_dir = os.getcwd() + "/src/simmer/tests/PHARO_image_driver/"
            delete_folder(raw_dir)
            self.assertTrue(val)
        except:
            e = sys.exc_info()[0]
            print(e)
            delete_folder(raw_dir)
            self.assertTrue(False)



"""
This module provides utility functions for the reduction pipeline.
"""

import astropy.io.fits as pyfits
import numpy as np


def find_angle(loc1, loc2):
    """
    Calculated the angle between two locations on a grid.

    Inputs:
        :loc1: (tuple) first location.
        :relative: (tuple) second location.

    Outputs:
        :angle : (float) real-valued angle between loc1 and loc2.
    """
    angle = np.atan(loc1[1] / loc2[1])
    return angle


def make_filelist(directory, numlist, inst):
    """Turn a list of numbers into a list of properly formatted filenames.

    Inputs:
        :directory: (string) path leading to directory of interest.
        :numlist: (list) list of numbers corresponding to fits files.
        :inst: (Instrument object) instrument for which data is being reduced.

    Outputs:
        :filelist: (list) list of strings pertaining to files of interest
    """

    filelist = [
        directory + inst.file_prefix + "{:04d}.fits".format(d) for d in numlist
    ]
    return filelist


def read_imcube(filelist):
    """Reads a stack of fits files into an image cube of dimensions (nims, xpix, ypix).

    Inputs:
        :filelist: (list) list of strings pertaining to files of interest.

    Outputs:
        :im_array: (3D array) array of 2D arrays pertaining to the files in filelist.

    """

    im_array = np.array([pyfits.getdata(file, 0) for file in filelist])

    return im_array


def image_subsection(input_image, npix, center):
    """reads in a full image array, selects the relevant subsection of the array,
    and returns the new array, transposed for use with Python.
    input_image must be 2D

    Inputs:
        :input_image: (2D array) image of which a subsection is desired.
        :center: (tuple) center of image, with format (x, y)
        :npix: (float) value for size of return image. If non-square image desired, enter as list.

        :default: center = (750, 1100)
        :npix: = 600 for inscribed region or npix = 1000 for circumscribing region

        :transposed: = np.rot90(input_image.T,2)
        :transposed: = np.rot90(input_image,2)
        :center: = (2047-center[0],2047-center[1]

    Outputs:
        :subsection: (2d array) subsection of original image.

    )"""

    npix = np.array(npix)
    if np.size(npix) == 1:
        npix = [npix, npix]
    half = [int(n / 2) for n in npix]

    subsection = input_image[
        center[0] - half[0] : center[0] + half[0],
        center[1] - half[1] : center[1] + half[1],
    ]

    return subsection


def header_subsection(input_image_file, npix, center):
    """
    Reads out the header of a subsection.

    Inputs:
        :input_image: (2D array) image of which a subsection is desired.
        :center: (tuple) center of image, with format (x, y)
        :npix: (float) value for size of return image. If non-square image desired, enter as list.

    Outputs:
        :header: (FITS header) header of the image file, adjusted accordingly.


    """
    header = pyfits.getheader(input_image_file)
    # hdulist = fits.open(input_image_file)
    # header = wcs.WCS(hdulist[0].header)

    header["CRPIX1"] = npix / 2 - (center[1] - header["CRPIX1"])  # x=col
    header["CRPIX2"] = npix / 2 - (center[0] - header["CRPIX2"])  # y=row
    header["NAXIS1"] = 600
    header["NAXIS2"] = 600

    return header


# def general_bad_pix(image):
#    sh = np.shape(image)
#    bp_im = image.copy()
#    px = 5
#
#    for r in range(sh[0]):
#        for c in range(sh[1]):
#            left = np.max([0, c-px]) #left of image, or 5 less than current pixel
#            right = np.min([sh[1], c+px]) #right of image or 5 more than current pixel
#            bott = np.max([0, r-px]) #bottom of image or 5 less than current pixel
#            top = np.min([sh[0], c+px]) #top of image or 5 more than current pixel
#
#            region = image[bott:top, left:right]
#            region_size = np.size(region)
#
#            nans = np.sum(np.isnan(region))
#            if nans == region_size:
#                #all these pixels are shitty, set value to 0
#                bp_im[r,c] = 0.
#            else:
#                r_med = np.nanmedian(region)
#                if image[r,c] > 5.*r_med or np.isnan(image[r,c]):
#                    bp_im[r,c] = r_med
#
#    return bp_im



"""
For ShARCS, the darks are often produced by an automated script at
the end of a night. This module adds these frames to the log sheet.
"""


# AS
# Created: 2/11/19
# Updated: 2/11/19

import glob
import os
from ast import literal_eval

import astropy.io.fits as pyfits
import numpy as np
import pandas as pd
from openpyxl import load_workbook


def add_dark_exp(inst, log, raw_dir, tab=None):
    """Adds dark exposures to the end of log sheet if not specified.

    Inputs:
        :tab: (string) tab of Excel sheet to be used.
        :inst: (Instrument object) instrument for which data is being reduced.
        :log: (string path of the logsheet.
        :raw_dir: (string) path of the directory containing the raw data.
    """

    def get_number(file):
        """For a Shane FITS filename, returns a number given its filename.

        Inputs:
            :file: (string) name of file of interest.
        """
        if file[1] == "0":
            number = literal_eval(file[2:5])  # just a safer eval
        print(file)
        number = literal_eval(file[1:5])
        return number

    def find_end(column):
        """Finds the end of a column.

        Inputs:
            :column: (pandas Series) column of DataFrame being searched.

        Outputs:
            :end: (int) index of end of column
        """
        for i, item in enumerate(column):
            if not isinstance(item, str) and not isinstance(item, int):
                end = i
                return end
        end = len(column)
        return len(column)

    def log_to_csv(log, tab, end, new_frame):
        """Writes a log to a csv file in current directory.

        Inputs:
            :log: (Excel sheet) logsheet to be analyzed.
            :tab: (string) tab of Excel sheet to be used.
            :end: (int) previous end of columns.
            :new_frame: (pandas DataFrame) new frame to be written into.
        """
        if tab:
            book = load_workbook(log)
            writer = pd.ExcelWriter(log, engine="openpyxl")
            writer.book = book
            writer.sheets = dict((ws.title, ws) for ws in book.worksheets)
            new_frame.to_excel(
                writer, tab, index=False, startrow=end, header=False
            )
            writer.save()
        else:
            os.remove(log)
            new_frame.to_csv(log)

    find_itimes(inst, raw_dir)
    if tab:
        initial_frame = pd.read_excel(pd.ExcelFile(log), tab)
    else:
        initial_frame = pd.read_csv(log)
    # testing that writing works well.
    # then going to create a new dataframe for the darks and append it.

    itimes_file = pd.read_csv(
        raw_dir + "_dark_itimes.txt", sep=" ", header=None
    )
    itimes_file.columns = ["file_name", "itime"]

    sorted_darks = itimes_file.sort_values(by=["file_name"])
    itimes = np.array(sorted_darks["itime"])
    files = np.array(sorted_darks["file_name"])
    starts = []
    ends = []
    objects = []
    exptime = []
    exposes = []

    for i, (itime, file) in enumerate(zip(itimes, files)):
        if itime not in exptime:
            exptime += [itime]
            starts += [get_number(file)]
            objects += ["dark"]
            if i != 0:
                end_file = files[i - 1]
                ends += [get_number(end_file)]
                exposes += [get_number(end_file) - starts[len(starts) - 2] + 1]
    end_file = files[len(files) - 1]
    ends += [get_number(end_file)]
    exposes += [get_number(end_file) - starts[len(starts) - 1] + 1]

    data_dict = {
        "Object": objects,
        "Start": starts,
        "End": ends,
        "ExpTime": exptime,
        "Coadds": np.full(len(objects), np.nan),
        "Expose": exposes,
        "Total_tint": np.full(len(objects), np.nan),
        "Filter": np.full(len(objects), np.nan),
        """Dither (")""": np.full(len(objects), np.nan),
        "Aperture": np.full(len(objects), np.nan),
        "TUB": np.full(len(objects), np.nan),
        "Airmass": np.full(len(objects), np.nan),
        "PT": np.full(len(objects), np.nan),
        "KepMag": np.full(len(objects), np.nan),
        "Companion": np.full(len(objects), np.nan),
        "Comments": np.full(len(objects), np.nan),
    }
    new_frame = pd.DataFrame(data=data_dict)

    end = find_end(initial_frame["Object"])
    if not end:
        end = len(initial_frame["Object"])

    log_to_csv(log, tab, end, new_frame)


def find_itimes(inst, raw_dir):
    """Read headers for all darks and make a list showing exposure times.

    Inputs:
        :inst: (Instrument object) instrument for which data is being reduced.
        :raw_dir: (string) path of the directory containing the raw data.

    """
    # Set directories
    outdir = raw_dir

    files = glob.glob(raw_dir + inst.file_prefix + "*.fits")
    #     lastitime = 0
    with open(outdir + "_dark_itimes.txt", "w") as outfile:
        for file in files:
            # load in header:
            head = pyfits.getheader(file)

            hkeys = np.array(list(head.keys()))

            # Strangely, there are one or two files without the "OBJECT" keyword
            # in the header. This check prevents crashes.
            integrations = np.where(hkeys == "TRUITIME")
            objects = np.where(hkeys == "OBJECT")
            if np.logical_and(
                len(hkeys[integrations]) > 0, len(hkeys[objects]) > 0
            ):
                itime = int(round(inst.itime(head)))
                obj = head["OBJECT  "]

                if obj == "dark":
                    #                     if lastitime != itime:
                    #                         lastitime = itime
                    outfile.write(
                        os.path.basename(file) + " " + str(itime) + "\n"
                    )



# -*- coding: utf-8 -*-

__uri__ = "https://simmer.readthedocs.io"
__author__ = "Arjun B. Savel, Lea Hirsch, Courtney D. Dressing, David R. Ciardi"
__email__ = "asavel@gmail.com"
__license__ = "MIT"
__version__ = "0.3.5"
__release__ = "0.3.5-beta"
__description__ = "Stellar Image Maturation via Efficient Reduction"



"""
Creates config file from logsheet. Will move into either utils or example folder.
"""

import pandas as pd


def create_config(log, config_file, tab=None):
    """
    Create config csv file out of tab in logsheet.

    Inputs:
        :log: (string) path of the logsheet.
        :config_file: (string) path of the desired concrete file, including the
                    file name.
        :tab: (string) tab of logsheet to be turned into.

    """
    if log[-3:] == "csv":
        logdf = pd.read_csv(log)
    elif log[-4:] == "xlsx" or log[-3:] == "xls":
        logdf = pd.read_excel(
            log,
            sheet_name=tab,
            header=0,
            parse_columns=11,
            converters={
                "Comments": str,
                "Start": int,
                "End": int,
                "Object": str,
                "Filter": str,
            },
        )
    logdf = logdf[pd.notna(logdf["Start"])]
    savedf = logdf[["Object", "ExpTime", "Filter", "Comments"]]

    nrows = len(logdf)

    filenums = []
    for row in range(0, nrows):
        start = logdf["Start"].iloc[row]
        end = logdf["End"].iloc[row]
        filelist = range(start, end + 1)
        filenums.append(filelist)

    savedf["Filenums"] = pd.Series(filenums)

    savedf.to_csv(config_file, index=False)



"""
Module containing all functions related to image center-finding
and stacking.
"""

import matplotlib.pylab as plt
import numpy as np
from scipy.ndimage.filters import median_filter
from scipy.ndimage.interpolation import rotate
from scipy.ndimage.interpolation import shift as subpix_shift
from skimage.feature import peak_local_max

from .scipy_utils import *


def roll_shift(image, shifts, cval=0.0):
    """Enter shifts as (drow, dcol)"""
    first_roll = np.roll(image, shifts[0], axis=0)
    if shifts[0] >= 0:
        first_roll[0 : shifts[0], :] = cval
    else:
        first_roll[shifts[0] :, :] = cval

    second_roll = np.roll(first_roll, shifts[1], axis=1)
    if shifts[1] >= 0:
        second_roll[:, 0 : shifts[1]] = cval
    else:
        second_roll[:, shifts[1] :] = cval

    return second_roll


def register_bruteforce(image, rough_center=None):
    """
    Performs the default image registration scheme. Shifts the center of the
    image to the peak.

    inputs:
        :image: (2-d array) photon counts at each pixel of each science image.
        :rough_center: (2-d array, default None) location of primary star. This
                    argument is only passed in the wide binary case.

    outputs:
        image_centered : (2-d array) image cenered by the rotations method.
    """

    def search_threshold(image):
        """
        Performs a binary search along local max thresholds. Returns coordinates corresponding
        to a threshold that only returns, at most, 3 peaks in the image.

        inputs:
            :image: (2d array) image data to be searched.

        outputs:
            :coordinates: (list) a m x 2 array

        """
        max_val = np.max(image)
        min_val = 0  # no negative values will be out peak

        # now perform binary search; first initialize lower, upper bounds
        lower_bound = min_val
        upper_bound = max_val
        while lower_bound <= upper_bound:
            threshold = np.floor((lower_bound + upper_bound) / 2)
            coordinates = peak_local_max(
                image, min_distance=100, threshold_abs=threshold
            )
            if len(coordinates) > 3:
                lower_bound = threshold + 1
            elif len(coordinates) == 0:
                upper_bound = threshold - 1
            else:
                return coordinates
        return []

    im_shape = np.shape(image)
    cent = (im_shape[0] / 2, im_shape[1] / 2)
    if rough_center:
        scale = np.floor(im_shape[0] / 50)
        small_image = image[
            rough_center[0] - scale : rough_center[0] + scale,
            rough_center[1] - scale : rough_center[1] + scale,
        ]
        # now do binary search
        small_coordinates = search_threshold(small_image)
        coordinates_y = [
            coord[0] + (rough_center[0] - scale) for coord in small_coordinates
        ]
        coordinates_x = [
            coord[1] + (rough_center[1] - scale) for coord in small_coordinates
        ]
        coordinates = tuple(zip(coordinates_y, coordinates_x))
    else:
        coordinates = search_threshold(image)
    num_rows, num_cols = np.shape(image)
    cent = (int(num_rows / 2), int(num_cols / 2))
    base_position = cent  # if no other information available, use center
    if len(coordinates) == 0:  # if the algorithm failed
        return []  # pass that along

    yshift = base_position[0] - coordinates[0][0]
    xshift = base_position[1] - coordinates[0][1]

    image_centered = roll_shift(image, (yshift, xshift))
    return image_centered


def find_wide_binary(image):
    """
    Performs the first step of image registration for a science image that
    contains a wide binary. User input selects which target is the primary star
    of interest in the first frame of the target.

    inputs:
        :image: (2-d array) photon counts at each pixel of each science image.

    outputs:
        :image_centered: (2-d array) image cenered by the rotations method.
    """

    def onclick(event):
        click_x, click_y = event.xdata, event.ydata
        rough_center.append((click_y, click_x))

        # Only select one star
        if len(rough_center) == 1:
            fig.canvas.mpl_disconnect(cid)
            plt.close(1)

    fig = plt.figure(1)
    ax = fig.add_subplot(111)
    ax.imshow(image)

    cid = fig.canvas.mpl_connect("button_press_event", onclick)

    rough_center = []
    plt.show(1)

    # fig.canvas.mpl_connect('button_press_event', onclick)
    return rough_center


def register_saturated(image, searchsize1, newshifts1, rough_center=None):

    """
    Performs image registration when a saturated star is present in
    the science image.

    inputs:
        :image: (2-d array) photon counts at each pixel of each science image.
        :searchsize1: (int) initial size of search for center of image.
        :newshifts1: (list) keeps tracks of x-y shifts.
        :rough_center: (2-d array, default None) location of primary star. This
                    argument is only passed in the wide binary case.

    outputs:
        :image_centered: (2-d array) image centered by the rotations method.
        :rot: the rotation
        :newshifts1: (list) keeps tracks of x-y shifts.
    """
    im_shape = np.shape(image)
    cent = (im_shape[0] / 2, im_shape[1] / 2)
    if rough_center:
        scale = np.floor(im_shape[0] / 50)
        image = image[
            rough_center[0] - scale : rough_center[0] + scale,
            rough_center[1] - scale : rough_center[1] + scale,
        ]

    res1, im1, (xshift1, yshift1) = run_rot(image, searchsize1, cent, 200)
    if np.max(res1) == 0:
        rot = np.empty(res1.shape)
        rot.fill(np.nan)
    else:
        rot = res1 / np.max(res1)
    newshifts1.append((yshift1, xshift1))
    image_centered = subpix_shift(image, (yshift1, xshift1))
    return image_centered, rot, newshifts1


def rot_search(dat, x_initial, y_initial, xrad, yrad):
    """
    Perform rotational search of an image.

    Inputs:
        :dat: (2d array) image data.
        :x_initial: (int) the initial guess for the x-coordinate of the center.
        :y_initial: (int) the initial guess for the y-coordinate of the center.
        :xrad: (int) radius in x to search.
        :yrad: (int) radius in y to search.

    outputs:
        :(xshift, yshift): (tuple) record of how much the image was shifted in x a dny
        :out: (2d array) output shifted image
    """
    # calculate offset from center
    xoffset = dat.shape[1] / 2 - x_initial  # from center of x
    yoffset = dat.shape[0] / 2 - y_initial
    # calcualte roll shifts for all x and y combinations within the
    # xrad, yrad regimes from the offset
    x_grid, y_grid = np.meshgrid(
        np.arange(xoffset + xrad, xoffset - xrad - 1, -1),
        np.arange(yoffset + yrad, yoffset - yrad - 1, -1),
    )

    out = []
    for (xshift, yshift) in zip(x_grid.flatten(), y_grid.flatten()):
        rolled = roll2d(dat, xshift, yshift)
        tot = rotate_sub(rolled)
        out.append(tot)
    out = np.array(out)
    out = np.reshape(out, (2 * yrad + 1, 2 * xrad + 1))

    pix = np.unravel_index(np.argmin(out), out.shape)
    xshift = x_grid[pix]
    yshift = y_grid[pix]

    return (xshift, yshift), out


def roll2d(dat, xshift, yshift):
    """
    Essentially performs numpy roll function in 2 dimensions.

    inputs:
        :dat: (2d array) image data.
        :xshift: (int) shift in the x direction.
        :yshift: (int) shift in the y direction.
    """
    # do x first
    #     for d in dat:
    #         dat = list(map(int, d))
    xshift = int(xshift)
    yshift = int(yshift)
    dat2 = np.roll(dat, xshift, axis=1)
    # zero out the wrapped around values
    if xshift > 0:
        dat2[:, 0:xshift] = 0
    else:
        dat2[
            :, dat.shape[1] + xshift : dat.shape[1]
        ] = 0  # plus sign because xshift is negative
    # now do y in similar fashion
    dat2 = np.roll(dat2, yshift, axis=0)
    if yshift > 0:
        dat2[0:yshift, :] = 0
    else:
        dat2[dat.shape[0] + yshift : dat.shape[0], :] = 0
    return dat2


def rotate_sub(dat):
    """Rotate the image about the center point, then subtract from original and record residuals.
    Does this at set angles.

    inputs:
        :dat: (2d array) image data.

    outputs:
        :total_residuals: (float) the summed total residuals.

    """
    angles = [90, 180, 270]
    total_residuals = 0.0

    for angle in angles:
        rotated_dat = rotate(dat, angle, reshape=False)
        residuals = np.sum(np.abs(dat - rotated_dat))
        total_residuals += residuals
    return total_residuals


def calc_shifts(
    dat, x_initial, y_initial, xrad, yrad, find="max", method="radon"
):
    """Do the radon search and then translate back to image coordinates."""

    if method == "rotate":
        out = rot_search(dat, x_initial, y_initial, xrad, yrad)[1]
    # elif method == 'radon':
    #     out = radonSearch(dat, x0, y0, xrad, yrad)

    # interpolate
    interped_out = imresize(out)
    # im = Image.fromarray(out) # trying PIL
    # size = tuple((np.array(im.size) * 100.).astype(int))
    # interped_out = np.array(im.resize(size, Image.BICUBIC))
    if find == "max":
        pix = np.unravel_index(np.argmax(interped_out), interped_out.shape)
    elif find == "min":
        pix = np.unravel_index(np.argmin(interped_out), interped_out.shape)

    xcen = dat.shape[1] / 2
    ycen = dat.shape[0] / 2
    xoffset = xcen - x_initial
    yoffset = ycen - y_initial
    # calcualte roll shifts for all x and y combinations
    x_grid, y_grid = np.meshgrid(
        np.arange(xoffset + xrad, xoffset - xrad - 1, -0.01),
        np.arange(yoffset + yrad, yoffset - yrad - 1, -0.01),
    )

    xshift = x_grid[pix]
    yshift = y_grid[pix]

    return (xshift, yshift), out


def shift_bruteforce(image, base_position=None):
    """This will shift the maximum pixel to base_position (i.e. the center of image).
    Make sure base_position is entered as (int,int)."""

    num_rows, num_cols = np.shape(image)
    cent = (int(num_rows / 2), int(num_cols / 2))
    if not base_position:
        base_position = cent  # if no other information available, use center

    # First, do a median filter on the image and find max pixel location.
    # Filter to remove hot pixels and make sure max is star.
    filtered = median_filter(image, size=7)
    maxpix = np.unravel_index(np.nanargmax(filtered), filtered.shape)

    # Now shift that location to the center (or base_position)
    yshift = base_position[0] - maxpix[0]
    xshift = base_position[1] - maxpix[1]

    shifted = roll_shift(image, (yshift, xshift))

    return shifted, (yshift, xshift)


def run_rot(image, searchsize, center, newsize):
    """
    Runs all rotations.
    """
    image[np.where(image < 0.0)] = 0.0
    cut_image = image[
        int(center[0] - newsize / 2) : int(center[0] + newsize / 2),
        int(center[1] - newsize / 2) : int(center[1] + newsize / 2),
    ]
    newcent = (newsize / 2, newsize / 2)

    # (xshift, yshift), res = r.rotSearch(im, newcent[0], newcent[1], searchsize, searchsize)
    (xshift, yshift), res = calc_shifts(
        cut_image,
        newcent[0],
        newcent[1],
        searchsize,
        searchsize,
        find="min",
        method="rotate",
    )

    return res, cut_image, (xshift, yshift)



import os

import cerberus
import yaml

from .custom_validator import SimmerValidator


def normalize(yml_dict, validator, schema, plot_types):
    """
    Inputs:
        :yml_dict: (dictionary) the dictionary to be normalized against a schema
        :validator: (SimmerValidator) the validator object used.
        :schema: the schema against which the yml_dict is normalized.
        :plot_types: (list of strings) the basic plot_types that must be in
            the uppermost keys.

    Outputs:
        :normalized: normalized dictionary.
    """
    validator.schema = schema
    for plot_type in plot_types:
        if plot_type not in yml_dict.keys():
            yml_dict[plot_type] = {}
    normalized = validator.normalized(yml_dict)
    return normalized


def read_yml(yml_filename):
    """
    Reads in a yaml file.

    Inputs:
        :yml_filename: (string) path to the yml.

    Outputs:
        :parsed_yml_file: (dictionary) key-value pairs as read from the yaml
            file.

    """
    file = open(yml_filename)
    parsed_yaml_file = yaml.load(file, Loader=yaml.SafeLoader)
    file.close()
    return parsed_yaml_file


def validate_yml(schema_filename, yml_filename):
    """
    Ensures that a given yml file is in accordance with the provided schema. In
    essence, this ensures that no odd keys or fields are provided to the yml.

    Inputs:
        :schema_filename: (string) path to schema yaml.
        :yml_filename: (string) path to yml yaml.

    Outputs:
        :validated: (bool) whether or not the yaml was successfully validated.

    """
    parsed_schema = read_yml(schema_filename)
    parsed_yml = read_yml(yml_filename)
    s = SimmerValidator()
    try:
        s.validate(parsed_yml, parsed_schema)
        validated = True
    except cerberus.SchemaError:
        validated = False
    return validated


def get_plotting_args(yml_filename=None):
    """
    Gets plotting args.

    Inputs:
        :yml_filename: (string) path of the plotting yml to be used.
                        Defaults to None.

    Outputs:
        :plotting_arg: (dictionary) all arguments that are related to plotting.
            See the `plotting_schema.yml` schema for documentation of keys and values.

    """
    my_path = os.path.abspath(os.path.dirname(__file__))
    schema_filename = os.path.join(my_path, "plotting_schema.yml")
    plot_types = ["intermediate", "final_im", "rots"]

    if not yml_filename:
        # the normalizer fills in all empty fields later on
        yml_dict = {plot_type: [{}] for plot_type in plot_types}
    else:
        if validate_yml(schema_filename, yml_filename):
            yml_dict = read_yml(yml_filename)
        else:
            raise cerberus.SchemaError("parsing plotting yml failed")

    s = SimmerValidator()
    schema = read_yml(schema_filename)
    plotting_args = normalize(yml_dict, s, schema, plot_types)
    return plotting_args



"""
    isort:skip_file
"""

import yaml
import unittest
from shutil import copyfile
import os

import pandas as pd
import numpy as np
import astropy.io.fits as pyfits
import simmer.add_dark_exp as ad
import simmer.schemas.custom_validator as validator
import simmer.schemas.read_yml as read
import simmer.insts as i
import simmer.create_config as c
import simmer.check_logsheet as check
import simmer.search_headers as search
from simmer.tests.tests_reduction import (
    download_folder,
    delete_folder,
    DataDownloadException,
)


class TestYml(unittest.TestCase):
    default_config = {
        "intermediate": [
            {
                "plot": True,
                "zoom_scale": 0,
                "colorbars": True,
                "colormap": "plasma",
                "scaling": "linear",
            }
        ],
        "final_im": [
            {
                "plot": True,
                "zoom_scale": 0,
                "colorbars": True,
                "colormap": "plasma",
                "scaling": "linear",
            }
        ],
        "rots": [
            {
                "plot": True,
                "zoom_scale": 0,
                "colorbars": True,
                "colormap": "plasma",
                "scaling": "linear",
            }
        ],
    }

    plot_types = ["intermediate", "final_im", "rots"]

    def test_validator_instantiate(self):
        try:
            validator.SimmerValidator()
            instantiated = True
        except:
            instantiated = False
        self.assertTrue(instantiated)

    def test_validate(self):
        schema_filename = "src/simmer/schemas/plotting_schema.yml"
        yml_filename = "src/simmer/tests/test_validate.yml"
        validated = read.validate_yml(schema_filename, yml_filename)
        self.assertTrue(validated)

    def test_empty_yaml(self):
        yml_dict = {"intermediate": [{}], "final_im": [{}], "rots": [{}]}
        ss = validator.SimmerValidator()
        file = open("src/simmer/schemas/plotting_schema.yml")
        parsed_yaml_file = yaml.load(file, Loader=yaml.SafeLoader)
        ss.schema = parsed_yaml_file
        file.close()
        self.assertEqual(ss.normalized(yml_dict), self.default_config)

    def test_negative_zoom(self):
        s = validator.SimmerValidator()

        file = open("src/simmer/schemas/plotting_schema.yml")
        parsed_yaml_file = yaml.load(file, Loader=yaml.SafeLoader)

        document = {"final_im": [{"zoom_scale": -10}]}
        validated = s.validate(document, parsed_yaml_file)
        file.close()
        self.assertFalse(validated)

    def test_wrong_type_zoom(self):
        s = validator.SimmerValidator()

        file = open("src/simmer/schemas/plotting_schema.yml")
        parsed_yaml_file = yaml.load(file, Loader=yaml.SafeLoader)

        document = {"final_im": [{"zoom_scale": "some_string"}]}
        validated = s.validate(document, parsed_yaml_file)
        file.close()
        self.assertFalse(validated)

    def test_colormap_validator_valid(self):
        ss = validator.SimmerValidator()
        file = open("src/simmer/schemas/plotting_schema.yml")
        parsed_yaml_file = yaml.load(file, Loader=yaml.SafeLoader)
        document = {"final_im": [{"colormap": "viridis"}]}
        validated = ss.validate(document, parsed_yaml_file)
        file.close()
        self.assertTrue(validated)

    def test_colormap_validator_invalid(self):
        s = validator.SimmerValidator()
        file = open("src/simmer/schemas/plotting_schema.yml")
        parsed_yaml_file = yaml.load(file, Loader=yaml.SafeLoader)
        document = {"final_im": [{"colormap": "badmap"}]}
        try:
            validated = s.validate(document, parsed_yaml_file)
        except ValueError:
            validated = False
        file.close()
        self.assertFalse(validated)


class TestConfig(unittest.TestCase):
    def test_create_config_csv(self):
        try:
            download_folder("config_test")
        except:
            raise DataDownloadException(
                "Could not download test data for config_test."
            )
        tab = "Sheet1"
        excel_path = "src/simmer/Examples/PHARO/logsheet.csv"
        c.create_config(excel_path, "created_frame.csv", tab)
        created_frame = pd.read_csv("created_frame.csv")
        compare_frame = pd.read_csv(
            "src/simmer/tests/config_test/compare_frame_csv.csv"
        )
        delete_folder("src/simmer/tests/config_test")
        self.assertTrue(created_frame.equals(compare_frame))

    def test_create_config_xlsx(self):
        try:
            download_folder("config_test")
        except:
            raise DataDownloadException(
                "Could not download test data for config_test."
            )
        tab = "Sheet1"
        excel_path = "src/simmer/Examples/Shane/logsheet.xlsx"
        c.create_config(excel_path, "created_frame.csv", tab)
        created_frame = pd.read_csv("created_frame.csv")
        compare_frame = pd.read_csv(
            "src/simmer/tests/config_test/created_frame_xlsx.csv"
        )
        delete_folder("src/simmer/tests/config_test")
        self.assertTrue(created_frame.equals(compare_frame))

    def test_check_logsheet_incorrect(self):
        inst = i.ShARCS()
        excel_path = "src/simmer/Examples/Shane/logsheet_incorrect.xlsx"
        failed = check.check_logsheet(inst, excel_path)
        self.assertTrue(failed == 7)

    def test_check_logsheet_correct(self):
        inst = i.ShARCS()
        excel_path = "src/simmer/Examples/PHARO/logsheet.csv"
        failed = check.check_logsheet(inst, excel_path)
        self.assertTrue(failed == 0)

    def test_add_dark_exposure(self):
        try:
            download_folder("dark_test")
        except:
            raise DataDownloadException(
                "Could not download test data for darks."
            )
        excel_path = "src/simmer/Examples/Shane/logsheet.xlsx"
        copy_path = "src/simmer/Examples/Shane/logsheet_copy.xlsx"
        copyfile(excel_path, copy_path)
        raw_dir = "src/simmer/tests/dark_test/"
        tab = "Sheet1"
        inst = i.ShARCS()
        ad.add_dark_exp(inst, copy_path, raw_dir, tab)
        os.remove(copy_path)
        delete_folder(raw_dir)
        self.assertTrue(True)


class TestPHAROSpecific(unittest.TestCase):
    inst = i.PHARO()

    def test_readpharo_quadrants(self):
        try:
            download_folder("readpharo_test")
        except:
            raise DataDownloadException(
                "Could not download test data for readpharo function."
            )
        raw_dir, new_dir, test_red_dir = (
            "src/simmer/tests/readpharo_test/raw/",
            "src/simmer/tests/readpharo_test/test_newdir/",
            "src/simmer/tests/readpharo_test/test_red/",
        )
        self.inst.read_data(raw_dir, new_dir)
        compare_flattened = pyfits.getdata(test_red_dir + "sph0436.fits")
        zero = np.zeros(np.shape(compare_flattened))
        flattened = pyfits.getdata(new_dir + "sph0436.fits")
        val = np.all(np.allclose(compare_flattened, flattened, equal_nan=True))
        delete_folder("src/simmer/tests/readpharo_test")
        self.assertTrue(val)


class TestShARCSSpecific(unittest.TestCase):
    inst = i.ShARCS()

    def test_search_headers(self):
        try:
            download_folder("sky_test")
        except:
            raise DataDownloadException(
                "Could not download test data for skies."
            )

        raw_dir, write_dir = (
            "src/simmer/tests/sky_test/",
            "src/simmer/tests/sky_test/",
        )
        search.search_headers(raw_dir, write_dir)
        val = True
        f = open(write_dir + "headers_wrong.txt", "r")
        for line in f:
            if "INCOMPLETE" in line:
                val = False
        f.close()
        self.assertTrue(val)



import matplotlib.cm as cm
from cerberus import Validator


class SimmerValidator(Validator):
    def _validate_type_mpl(self, value):
        """ Test the oddity of a value.

        The rule's arguments are validated against this schema:
        {'type': 'string'}
        """
        try:
            cm.get_cmap(value)
            return True
        except ValueError:  # if the colormap isn't valid
            return False



import matplotlib.colors as co
import matplotlib.pyplot as plt
import numpy as np

from .schemas import read_yml as read

plot_config = None


def initialize_plotting(yml_filename=None):
    global plot_config
    plot_config = read.get_plotting_args(
        yml_filename
    )  # call with empty arguments


def check_plot_type(plot_type):
    """
    Ensures that invalid plot types don't make it to plotting functions.

    Inputs:
        plot_type: (string) type of plot to be used.
    """
    valid_plot_types = ["rots", "final_im", "intermediate"]
    if plot_type not in valid_plot_types:
        raise NotImplementedError(
            "Plotting is not implemented for this plot type."
        )


def zoom(image, zoom_scale):
    """
    TODO: Check that this works for odd, even zoom_scale
    TODO: write tests

    Zooms in on center of image, with a representative zoom_scale.

    Inputs:
        :image: (numpy array) image to be zoomed in on.
        :zoom_scale: (int) length of side of box of zoomed image, in pixels.

    Outputs:
        :zoomed: (numpy array) zoomed image.
    """
    cent_x = im_shape[0] / 2
    cent_y = im_shape[1] / 2
    zoomed = image[cent_x - zoom_scale / 2 : cent_x + zoom_scale / 2][
        cent_y - zoom_scale / 2 : cent_y + zoom_scale / 2
    ]
    return zoomed


def add_colorbars(fig, plot_type, cim):
    """
    Inputs:
        fig: (figure object) This is where the colorbars are added.
        plot_type: (string) The type of plot to add colorbars to.
    """
    scaling = plot_config[plot_type][0]["scaling"]
    text_dict = {
        "rots": "Residuals",
        "final_im": "Counts",
        "intermediate": "Counts",
    }
    cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
    cbar = fig.colorbar(cim, cax=cbar_ax)
    text = f"{text_dict[plot_type]}, {scaling} scaling"
    cbar.set_label(text)
    cbar.ax.tick_params(labelsize=50)


def plot_array(
    plot_type, im_array, vmin, vmax, directory, filename, extent=None
):  # pylint: disable=too-many-arguments
    """
    Plots arrays produced in the process of the pipeline.

    Inputs:
        :im_array: (3D array) array of 2D images.
        :vmin: (int) minimum for linear color mapping of plots.
        :vmax: (int) maximum for linear color mapping of plots.
        :directory: (str) path to directory to which the image file will be saved.
        :filename: (str) name of file to which the image will be saved.
        :extent: (tuple) from matplotlib documentation: controls the bounding box in data coordinates that
                                the image will fill specified as (left, right, bottom, top) in data coordinates

    Outputs:
        :fig: (Matplotlib figure) plotted figure.
    """

    def plot_few(func):
        fig = plt.figure(figsize=(30, 6))
        for i in range(array_len):
            ax = fig.add_subplot(
                1, array_len, i + 1
            )  # pylint: disable=invalid-name # common axis name!
            pltim = np.rot90(im_array[i, :, :], 2)
            scaled_pltim = func(pltim)

            cim = ax.imshow(
                scaled_pltim,
                origin="lower",
                cmap="plasma",
                norm=co.Normalize(vmin=vmin, vmax=vmax),
                extent=extent,
            )
            ax.tick_params(axis="both", which="major", labelsize=20)
        return fig, cim

    def plot_many(func):
        nrows = 4
        ncols = int(np.ceil((array_len / 4.0)))
        rowheight = nrows * 10
        colheight = ncols * 10

        fig = plt.figure(figsize=(colheight, rowheight))
        for i in range(array_len):
            ax = fig.add_subplot(
                nrows, ncols, i + 1
            )  # pylint: disable=invalid-name # common axis name!
            if filename == "centers.png":
                pltim = np.rot90(im_array[i, 250:350, 250:350], 2)
            else:
                pltim = np.rot90(im_array[i, :, :], 2)
            scaled_pltim = func(pltim)

            cim = ax.imshow(
                scaled_pltim,
                origin="lower",
                cmap=plot_config[plot_type][0]["colormap"],
                norm=co.Normalize(vmin=vmin, vmax=vmax),
                extent=extent,
            )
            ax.tick_params(axis="both", which="major", labelsize=40)
            if filename == "centers.png":
                ax.plot([50], [50], "wo", markersize=8)
                ax.set_xlim([0, 100])
                ax.set_ylim([0, 100])
        return fig, cim

    if not plot_config:
        initialize_plotting()

    # if this shouldn't be plotted, just return
    if not plot_config[plot_type][0]["plot"]:
        return
    check_plot_type(plot_type)
    func_dict = {
        "linear": lambda x: x,
        "sine": np.sin,
        "exponential": np.exp,
        "log": np.log10,
        "quadratic": np.square,
        "square root": np.sqrt,
    }

    func = func_dict[plot_config[plot_type][0]["scaling"]]

    array_len = np.shape(im_array)[0]

    if array_len <= 5:
        fig, cim = plot_few(func)

    elif array_len > 50:
        print("Too many images to plot.")
        return
    else:  # 11 images? 13 images? make it 4xn
        fig, cim = plot_many(func)

    fig.subplots_adjust(right=0.8)
    if plot_config[plot_type][0]["colorbars"]:
        add_colorbars(fig, plot_type, cim)
    plt.savefig(directory + filename)
    plt.close("all")
    return fig



"""
Module to make the image stacking / production occur.
"""


import os
from glob import glob

import astropy.io.fits as pyfits
import numpy as np
from tqdm import tqdm

from . import plotting as pl
from . import registration as reg
from . import utils as u


def image_driver(raw_dir, reddir, config, inst, plotting_yml=None):
    """Do flat division, sky subtraction, and initial alignment via coords in header.
    Returns Python list of each registration method used per star.

    Inputs:
        :raw_dir: (string) directory for the raw data
        :reddir: (string) directory for the reduced data
        :config: (pandas DataFrame) dataframe corresponding to config sheet for data.
        :inst: (Instrument object) instrument for which data is being reduced.
        :plot: (bool) determines whether or not intermediate plots should be produced.

    """
    # Save these images to the appropriate folder.

    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    if inst.take_skies:
        skies = config[config.Comments == "sky"]
    else:
        skies = config[
            (config.Object != "flat")
            & (config.Object != "dark")
            & (config.Object != "setup")
        ]
    stars = skies.Object.unique()
    sdirs = glob(reddir + "*/")

    methods = []

    for star in tqdm(
        stars, desc="Running image driver", position=0, leave=True
    ):
        s_dir = reddir + star + "/"
        if (
            s_dir not in sdirs
        ):  # make sure there's a subdirectory for each star
            os.mkdir(s_dir)

        filts = skies[
            skies.Object == star
        ].Filter.values  # array of filters as strings
        for n, filter_name in enumerate(filts):
            obj = config[config.Object == star]
            #             imlist = literal_eval(obj[obj.Comments != 'sky'].Filenums.values[n])
            imlist = eval(
                obj[obj.Comments != "sky"].Filenums.values[n]
            )  # pylint: disable=eval-used # liter_eval issues
            # cast obj_methods as list so that elementwise comparison isn't performed
            obj_methods = list(config[config.Object == star].Method.values)
            if "saturated" and "wide" in obj_methods:
                methods.append("saturated wide")
            elif "saturated" in obj_methods and "wide" not in obj_methods:
                methods.append("saturated")
            elif "saturated" not in obj_methods and "wide" in obj_methods:
                methods.append("wide")
            else:
                methods.append("default")
            create_imstack(
                raw_dir, reddir, s_dir, imlist, inst, filter_name=filter_name
            )
    return methods


def create_imstack(
    raw_dir, reddir, s_dir, imlist, inst, plotting_yml=None, filter_name=None
):
    """Create the stack of images by performing flat division, sky subtraction.

    Inputs:
        :raw_dir: (string) directory for the raw data
        :reddir: (string) directory for the reduced data
        :s_dir: (string) directory corresponding to a specific star.
        :imlist: (list) list of strings of paths pointing to image files.
        :inst: (Instrument object) instrument for which data is being reduced.
        :plot: (bool) determines whether or not intermediate plots should be produced.
        :fiilter_name: (string) name of the filter used for the images in question.

    Outputs:
        :im_array: (3d array) array of 2d images.
        :shifts_all: recording of all the x-y shifts made
    """
    #     method = 'brute force' # the default
    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    nims = len(imlist)
    imfiles = u.make_filelist(raw_dir, imlist, inst)

    im_array = u.read_imcube(imfiles)

    im_array = inst.adjust_array(im_array, nims)

    head = inst.head(imfiles[0])
    filt = inst.filt(nims, head, filter_name)

    # if necessary, make directory for filter. Also grab correct flat file
    fdirs = glob(s_dir + "*/")

    sf_dir = s_dir + filt + "/"
    if sf_dir not in fdirs:  # make a directory for each filt
        os.mkdir(sf_dir)

    flatfile = reddir + f"flat_{filt}.fits"
    if (
        inst.name == "PHARO" and filt == "Br-gamma"
    ):  # not sure whether this is generalizable
        flatfile = reddir + "flat_K_short.fits"
    flat = pyfits.getdata(flatfile, 0)

    skyfile = sf_dir + "sky.fits"
    sky = pyfits.getdata(skyfile, 0)
    sky[sky != sky] = 0.0  # set nans from flat=0 pixels to 0 in sky

    shifts_all = []
    for i in range(nims):
        # flat division and sky subtraction
        current_im = im_array[i, :, :]
        flat[flat == 0] = np.nan
        current_im = (
            current_im / flat
        ) - sky  # where flat = 0, this will be nan
        current_head = pyfits.getheader(imfiles[i])

        # bad pixel correction
        current_im = inst.bad_pix(current_im)

        # now deal with headers and shifts
        shifted_im, shifts = reg.shift_bruteforce(
            current_im
        )  # put it at the center

        shifts_all.append(shifts)

        #         if tests: # want to perform tests after dark subtraction and flat-fielding
        #             wides.append(reg.test_wide_binary)
        #             saturateds.append(reg.test_saturated)

        im_array[i, :, :] = shifted_im
        hdu = pyfits.PrimaryHDU(shifted_im, header=current_head)
        hdu.writeto(
            sf_dir + "sh{:02d}.fits".format(i),
            overwrite=True,
            output_verify="ignore",
        )

    pl.plot_array(
        "intermediate", im_array, -10.0, 10000.0, sf_dir, "shift1_cube.png"
    )

    # write shifts to file
    textfile = open(sf_dir + "shifts.txt", "w")
    textfile.write("im, d_row, d_col\n")
    for i, shift in enumerate(shifts_all):
        textfile.write("{},{},{}\n".format(i, *shift))
    textfile.close()
    return im_array, shifts_all


def create_im(s_dir, ssize1, plotting_yml=None, fdirs=None, method="default"):
    """Take the shifted, cut down images from before, then perform registration
    and combine. Tests should happen before this, as this is a per-star basis.
    """
    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    if not fdirs:
        fdirs = glob(s_dir + "*/")

    for sf_dir in fdirs:  # each filter

        # files = glob(sf_dir + 'sh*.fits') #reduced image files
        files = glob(
            sf_dir + f"s*.fits"
        )  # might need to change to file_prefix
        nims = len(files)
        frames = u.read_imcube(files)
        frames = frames.astype(float)

        arrsize1 = ssize1 * 2 + 1
        rots = np.zeros((nims, arrsize1, arrsize1))
        newshifts1 = []
        # newshifts2 = []
        for i in range(nims):  # each image
            image = frames[i, :, :]
            # image_centered, rot, newshifts1 = reg.register_saturated(image, ssize1, newshifts1)
            # rots[i, :, :] = rot

            if method == "saturated":
                image_centered, rot, newshifts1 = reg.register_saturated(
                    image, ssize1, newshifts1
                )
                rots[i, :, :] = rot
            elif method == "default":
                for j in range(len(image)):
                    for k in range(len(image[i])):  # change to j
                        if image[j][k] < 0.0:
                            image[j][k] = 0.0
                image_centered = reg.register_bruteforce(image)
                if len(image_centered) == 0:
                    print("Resorting to saturated mode.")
                    image_centered, rot, newshifts1 = reg.register_saturated(
                        image, ssize1, newshifts1
                    )
                    rots[i, :, :] = rot
            elif method == "saturated wide":
                rough_center = reg.find_wide_binary(image)
                image_centered = reg.register_bruteforce(
                    image, rough_center=rough_center
                )
            elif method == "wide":
                rough_center = reg.find_wide_binary(image)
                image_centered = reg.register_bruteforce(
                    image, rough_center=rough_center
                )
            frames[i, :, :] = image_centered  # newimage

        final_im = np.nanmedian(frames, axis=0)

        head = pyfits.getheader(files[0])
        hdu = pyfits.PrimaryHDU(final_im, header=head)
        hdu.writeto(
            sf_dir + "final_im.fits", overwrite=True, output_verify="ignore"
        )

        textfile1 = open(sf_dir + "shifts2.txt", "w")
        textfile1.write("im, d_row, d_col\n")
        for i, item in enumerate(newshifts1):
            textfile1.write("{},{},{}\n".format(i, *item))
        textfile1.close()
        pl.plot_array(
            "rots",
            rots,
            0.0,
            1.0,
            sf_dir,
            "rots.png",
            extent=[-ssize1, ssize1, -ssize1, ssize1],
        )
        pl.plot_array(
            "final_im", final_im, 0.0, 10000.0, sf_dir, "final_image.png"
        )
        pl.plot_array(
            "intermediate", frames, 0.0, 10000.0, sf_dir, "centers.png"
        )



"""
Module for driving large reduction processes.
"""

from glob import glob

import pandas as pd
from tqdm import tqdm

from . import darks, flats, image
from . import plotting as pl
from . import search_headers as search
from . import sky


def all_driver(inst, config_file, raw_dir, reddir, plotting_yml=None):
    """
    Runs all drivers, performing an end-to-end reduction.

    Inputs:
        :inst: (Instrument object) instrument for which data is being reduced.
        :config_file: (string) path of the config file containing plotting
            specifications. Optional.
        :plotting_yml: (string) path to the y
        :raw_dir: (string) path of the directory containing the raw data.
        :reddir: (string) path of the directory to contain the raw data.

    """

    # obtain file list from config file
    config = pd.read_csv(config_file)
    config.Object = config.Object.astype(str)

    if inst.name == "ShARCS":
        search.search_headers(raw_dir)

    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    darks.dark_driver(raw_dir, reddir, config, inst)
    flats.flat_driver(raw_dir, reddir, config, inst)
    sky.sky_driver(raw_dir, reddir, config, inst)
    methods = image.image_driver(raw_dir, reddir, config, inst)

    star_dirlist = glob(reddir + "*/")
    i = 0  # can't use tqdm on zipped objects, I believe
    for s_dir in tqdm(
        star_dirlist, desc="Running registration", position=0, leave=True
    ):
        image.create_im(s_dir, 10, method=methods[i])
        i += 1


def config_driver(inst, config_file, raw_dir, reddir):
    """
    Runs all_drivers, terminating afrer running sky_driver.

    Inputs:
        :inst: (Instrument object) instrument for which data is being reduced.
        :config_file: (string) path of the config file.
        :raw_dir: (string) path of the directory containing the raw data.
        :reddir: (string) path of the directory to contain the raw data.

    """

    # get file list from config file
    config = pd.read_csv(config_file)
    config.Object = config.Object.astype(str)

    darks.dark_driver(raw_dir, reddir, config, inst)
    flats.flat_driver(raw_dir, reddir, config, inst)
    sky.sky_driver(raw_dir, reddir, config, inst)


def image_driver(inst, config_file, raw_dir, reddir):
    """
    Runs all_drivers, terminating afrer running image_drivers.

    Inputs:
        :inst: (Instrument object) instrument for which data is being reduced.
        :config_file: (string) path of the config file.
        :raw_dir: (string) path of the directory containing the raw data.
        :reddir: (string) path of the directory to contain the raw data.
    """

    # get file list from config file
    config = pd.read_csv(config_file)
    config.Object = config.Object.astype(str)

    image.image_driver(raw_dir, reddir, config, inst)

    # Now do registration
    star_dirlist = glob(reddir + "*/")
    for s_dir in star_dirlist:
        image.create_im(s_dir, 10)



"""
Utility functions (used in the registration stage) adapted from SciPy 1.2.0 that have
since been deprecated (see the LICENSE.MD file at the project root directory).
"""

import numpy as np
from PIL import Image


def bytescale(data, high=255, low=0):
    """
    Byte scales an image.

    inputs:
        :data: (ndarray) PIL image data array.
        :cmin: (int)  Bias scaling of small values. Default is ``data.min()``.
        :cmax: (int) Bias scaling of large values. Default is ``data.max()``.
        :high: (int) Scale max value to `high`.  Default is 255.
        :low: (int) Scale min value to `low`.  Default is 0.

    outputs:
        img_array : (uint8 ndarray) The byte-scaled array.
    """
    cmin = data.min()
    cmax = data.max()

    cscale = cmax - cmin
    if cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data - cmin) * scale + low
    return (bytedata.clip(low, high) + 0.5).astype(np.uint8)


def toimage(arr, high=255, low=0):
    """Takes a numpy array and returns a PIL image.
    This function is only available if Python Imaging Library (PIL) is installed.
    The mode of the PIL image depends on the array shape and the `pal` and
    `mode` keywords.
    For 2-D arrays, if `pal` is a valid (N,3) byte-array giving the RGB values
    (from 0 to 255) then ``mode='P'``, otherwise ``mode='L'``, unless mode
    is given as 'F' or 'I' in which case a float and/or integer array is made.
    .. warning::
        This function uses `bytescale` under the hood to rescale images to use
        the full (0, 255) range if ``mode`` is one of ``None, 'L', 'P', 'l'``.
        It will also cast data for 2-D images to ``uint32`` for ``mode=None``
        (which is the default).
    Notes
    -----
    For 3-D arrays if one of the dimensions is 3, the mode is 'RGB'
    by default or 'YCbCr' if selected.
    The numpy array must be either 2 dimensional or 3 dimensional.
    """

    data = np.asarray(arr)
    shape = list(data.shape)
    valid = len(shape) == 2 or (
        (len(shape) == 3) and ((3 in shape) or (4 in shape))
    )
    if not valid:
        raise ValueError(
            "'arr' does not have a suitable array shape for " "any mode."
        )
    if len(shape) == 2:
        shape = (shape[1], shape[0])  # columns show up first
        bytedata = bytescale(data, high=high, low=low)
        image = Image.frombytes("L", shape, bytedata.tostring())
        return image


def imresize(arr):
    """
    Resizes an image.

    inputs
        :arr: (ndarray) The array of image to be resized.

    outputs:
        :im_resized: (ndarray) The resized array of image.
    """

    im = toimage(arr)
    size = tuple((np.array(im.size) * 100.0).astype(int))
    im_resized = im.resize(size, resample=3)
    im_resized = np.array(im_resized)
    return im_resized



"""
Functions to work with skies.
"""

import os
from glob import glob

import astropy.io.fits as pyfits
import numpy as np
from tqdm import tqdm

from . import plotting as pl
from . import utils as u

CENTER = (750, 1100)  # row,col
NPIX = 600


def sky_driver(raw_dir, reddir, config, inst, plotting_yml=None):
    """Night should be entered in format 'yyyy_mm_dd' as string.
    This will point toward a config file for the night with flats listed.

    Inputs:
        :raw_dir: (string) directory for the raw data
        :reddir: (string) directory for the reduced data
        :config: (pandas DataFrame) dataframe corresponding to config sheet for data.
        :inst: (Instrument object) instrument for which data is being reduced.
        :plot: (bool) determines whether or not intermediate plots should be produced.
    """

    if inst.take_skies:
        skies = config[config.Comments == "sky"]
    else:
        skies = config[
            (config.Object != "flat")
            & (config.Object != "dark")
            & (config.Object != "setup")
        ]

    stars = skies.Object.unique()
    sdirs = glob(reddir + "*/")

    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    for star in tqdm(stars, desc="Running skies", position=0, leave=True):

        s_dir = reddir + star + "/"
        if (
            s_dir not in sdirs
        ):  # make sure there's a subdirectory for each star
            os.mkdir(s_dir)

        filts = skies[
            skies.Object == star
        ].Filter.values  # array of filters as strings
        for n, filter_name in enumerate(filts):
            # CDD change
            # skylist = literal_eval(skies[skies.Object == s].Filenums.values[n])
            # literal_eval issues below
            skylist = eval(
                skies[skies.Object == star].Filenums.values[n]
            )  # pylint: disable=eval-used
            # end CDD change
            create_skies(
                raw_dir, reddir, s_dir, skylist, inst, filter_name=filter_name
            )


def create_skies(
    raw_dir, reddir, s_dir, skylist, inst, plotting_yml=None, filter_name=None
):
    """Create a sky from a single list of skies.
    sf_dir is the reduced directory for the specific star and filter.

    Inputs:
        :raw_dir: (string) directory for the raw data
        :reddir: (string) directory for the reduced data
        :s_dir: (string) directory corresponding to a specific star.
        :skylist: (list) list of strings of paths pointing to sky files.
        :inst: (Instrument object) instrument for which data is being reduced.
        :plot: (bool) determines whether or not intermediate plots should be produced.

    Outputs:
        :final_sky: (2D array) medianed sky image.
    """
    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    nskies = len(skylist)

    skyfiles = u.make_filelist(raw_dir, skylist, inst)

    sky_array = u.read_imcube(skyfiles)

    sky_array = inst.adjust_array(sky_array, nskies)

    head = inst.head(skyfiles[0])
    filt = inst.filt(nskies, head, filter_name)

    # if necessary, make directory for filter. Also grab correct flat file
    fdirs = glob(s_dir + "*/")

    sf_dir = s_dir + filt + "/"
    if sf_dir not in fdirs:  # make a dir for each filt
        os.mkdir(sf_dir)
    flatfile = reddir + f"flat_{filt}.fits"
    if inst.name == "PHARO" and filt == "Br-gamma":
        flatfile = reddir + "flat_K_short.fits"  # no BrG flats in Palomar data
    flat = pyfits.getdata(flatfile, 0)

    for i in range(nskies):
        flat[flat == 0] = np.nan
        sky_array[i, :, :] = sky_array[i, :, :] / flat

    # final_sky = u.median_outlier_reject(sky_array, 2.0) #2sigma outlier rejection?
    final_sky = np.median(sky_array, axis=0)

    pl.plot_array(
        "intermediate", sky_array, -10.0, 100.0, sf_dir, "sky_cube.png"
    )
    # CDD change
    #  head.update('DATAFILE', str(skylist)) #add all file names
    head.set("DATAFILE", str(skylist))  # add all file names
    # end CDD change

    hdu = pyfits.PrimaryHDU(final_sky, header=head)
    hdu.writeto(sf_dir + "sky.fits", overwrite=True, output_verify="ignore")

    return final_sky



"""Search for issues in FITS file headers similar
    to s0512.fits from night 2015_10_26, which did not
    contain either/both DATAFILE or/and FRAMENUM in
    the header. These issues are only known to occur
    with ShARCS data."""

from glob import glob

import astropy.io.fits as pyfits


def search_headers(raw_dir, write_dir=None):
    """Function to perform search of FITS headers.

    Inputs:
        :raw_dir: (string) absolute path to directory containing raw data.
        :write_dir: (string) absolute path to directory containing raw data.
                    Defaults to None; if this is the case, it's reassigned to
                    the `raw_dir` directory.

    """
    if not write_dir:
        write_dir = raw_dir

    file = write_dir + "headers_wrong.txt"
    textfile = open(file, "w")

    files = glob(raw_dir + "*.fits")
    files = [
        f
        for f in files
        if "sky" not in f.split("/")[-1]
        and "flat" not in f.split("/")[-1]
        and "dark" not in f.split("/")[-1]
    ]

    for file in files:
        # load in header
        head = pyfits.getheader(file)
        keys = head.keys()

        # check keywords
        key1 = "DATAFILE"
        key2 = "FRAMENUM"

        if key1 and key2 in keys:
            head_k1 = head[key1]
            head_k2 = head[key2]

            if int(head_k1[1:]) != head_k2:
                print(int(head_k1[1:]), head_k2, file)
                textfile.write(f"{file}\n")

        else:
            print(f"Header Incomplete in {file}!!! ")
            textfile.write(f"{file} HEADER INCOMPLETE\n")

    textfile.close()



"""
Functions to work with darks.
"""

import astropy.io.fits as pyfits
import numpy as np
from tqdm import tqdm

from . import plotting as pl
from . import utils as u

CENTER = (750, 1100)  # row,col
NPIX = 600


def dark_driver(raw_dir, reddir, config, inst, plotting_yml=None):
    """Night should be entered in format 'yyyy_mm_dd' as string.
    This will point toward a config file for the night with darks listed.flat

    Inputs:
        :raw_dir: (string) directory for the raw data
        :reddir: (string) directory for the reduced data
        :config: (pandas DataFrame) dataframe corresponding to config sheet for data.
        :inst: (Instrument object) instrument for which data is being reduced.
        :plot: (bool) determines whether or not intermediate plots should be produced.

    """
    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    _darks = config[config.Object == "dark"]
    texps = _darks.ExpTime.unique()

    for texp in tqdm(texps, desc="Running darks", position=0, leave=True):
        # literal_eval issues below
        darklist = eval(
            _darks[_darks.ExpTime == texp].Filenums.values[0]
        )  # pylint: disable=eval-used
        create_darks(
            raw_dir, reddir, darklist, inst
        )  # creates a new dark file


def create_darks(raw_dir, reddir, darklist, inst):
    """creates the actual darks from a list of dark file numbers, taking
    the median and writing to a file. Returns the final dark.

    Inputs:
        :reddir: (string) directory for the reduced data
        :darklist: string) list of dark file numbers
        :inst: (Instrument object) instrument for which data is being reduced.
    """

    ndarks = len(darklist)
    darkfiles = u.make_filelist(raw_dir, darklist, inst)

    dark_array = u.read_imcube(darkfiles)

    dark_array = inst.adjust_array(dark_array, ndarks)
    head = inst.head(darkfiles[0])
    itime = inst.itime(head)

    final_dark = np.median(dark_array, axis=0)  # nanmedian?

    pl.plot_array(
        "intermediate",
        dark_array,
        -1.0,
        50.0,
        reddir,
        f"dark_cube_{int(round(itime))}sec.png",
    )

    # CDD update
    # head.update('DATAFILE', str(darklist)) #add all file names
    head.set("DATAFILE", str(darklist))  # add all file names
    # end CDD update

    hdu = pyfits.PrimaryHDU(
        final_dark, header=head
    )  # can i do this with fits?
    dark_filename = reddir + "dark_{}sec.fits".format(int(round(itime)))
    hdu.writeto(dark_filename, overwrite=True, output_verify="ignore")

    return final_dark



"""
Module for instrument class and subclasses.
"""

import glob
import os

import astropy.io.fits as pyfits
import numpy as np
from scipy.ndimage.filters import median_filter

from . import utils as u


class Instrument:
    """
    Instantiates an object that dictates instrument-specific reduction techniques.
    """

    name = None

    npix = (
        10  # Size of rot search; needs to be bigger if initial shifts are off.
    )

    def __init__(self, take_skies=False):
        self.take_skies = take_skies

    def bad_pix(self, image):
        """Read in bp file, cut down to size, replace bad
        pixels with median of surrounding pixels.
        """
        c_im = image.copy()

        for i in range(3):
            filtered = median_filter(c_im, size=10)
            bad = np.where(c_im != c_im)  # check this
            c_im[bad] = filtered[bad]

        return c_im


class ShARCS(Instrument):
    """
    For use on the ShARCS camera on the Shane 3m telescope
    at Lick observatory.
    """

    name = "ShARCS"
    center = (750, 1100)  # row,col
    npix = 600

    plate_scale = 0.033  # arcsec/pixel

    replace_filters = {
        "BrG-2.16": ["Ks", "K"],
        "K+CH4-2.4": ["Ks", "K"],
        "J+CH4-1.2": "J",
        "H2-2.2122": ["Ks", "K"],
    }

    filter_logtohead = {
        "Ks": "Ks",
        "BrG": "BrG-2.16",
        "J+CH4-1.2": "J",
        "K": "Ks",
    }
    filter_headtolog = {"Ks": "K", "BrG-2.16": "BrG", "J": "J+CH4-1.2"}

    file_prefix = "s"
    plim_inds = (250, 350)
    off = (-250, -250)

    def adjust_array(self, array, nims):
        return np.array(
            [
                u.image_subsection(array[dd, :, :], self.npix, self.center)
                for dd in range(nims)
            ]
        )

    def adjust_im(self, image):
        return np.fliplr(image)

    def head(self, file):
        """
        Given a FITS file, returns its head.
        """
        return u.header_subsection(file, self.npix, self.center)

    def filt(self, nims, head, filter_name):
        """
        Given the header of a FITS file, returns its filter.
        """
        if head["FILT1NAM"] == "Unknown":
            filt = filter_name
        else:
            filt = head["FILT1NAM"]
        return filt

    def itime(self, head):
        """
        Given a FITS header, returns the true integration time
        for a file.
        """
        return head["ITIME0"] * 1e-6

    def bad_pix(self, image):
        """Read in bp file, cut down to size, replace bad
        pixels with median of surrounding pixels.
        """
        script_dir = os.path.dirname(
            __file__
        )  # <-- absolute dir the script is in
        rel_path = "badpix.fits"
        bpfile_name = os.path.join(script_dir, rel_path)
        bpfile = pyfits.getdata(bpfile_name, 0)
        bpfile = u.image_subsection(bpfile, self.npix, self.center)

        bad = np.where(bpfile == 1)  # locations of bad pixels

        filtered = median_filter(image, size=7)
        image[bad] = filtered[
            bad
        ]  # replace bad pixels with median of surrounding pixels

        return image

    def adjust_thisimage(self, thisimage, rawfile):
        thisimage = u.image_subsection(thisimage, self.npix, self.center)
        head = u.header_subsection(rawfile, self.npix, self.center)
        return thisimage, head

    def read_data(self, night, rawfilename, newfilename):
        raise NotImplementedError(
            "Data need not be read through"
            "this method for ShARCS. Instead,"
            "please run the driver function of "
            "your choice on folders containing "
            "your raw data."
        )


class PHARO(Instrument):
    """
    For use on the PHARO instrument at Palomar.
    """

    name = "PHARO"
    center = np.nan
    npix = np.nan  # Shouldn't matter
    plate_scale = 0.025
    filter_logtohead = {
        "Ks": "K_short",
        "BrG": "Br-gamma",
        "BrG+H2": "Br-gamma",
        "J": "J",
    }
    filter_headtolog = {"K-short": "Ks", "Br-gamma": "BrG", "J": "J"}
    replace_filters = {
        "BrG-2.16": ["Ks", "K"],
        "K+CH4-2.4": ["Ks", "K"],
        "J+CH4-1.2": "J",
        "H2-2.2122": ["Ks", "K"],
    }
    file_prefix = "sph"
    plim = (462, 562)
    off = (-462, -462)

    def adjust_im(self, image):
        return image

    def adjust_array(self, array, nims):
        return array.astype(float)

    def filt(self, nims, head, filter_name):
        """
        Given the header of a FITS file, returns its filter.
        """
        filt = head["FILTER"]
        return filt

    def head(self, file):
        """
        Returns the head of a FITS file.
        """
        return pyfits.getheader(file)

    def itime(self, head):
        """
        Given a FITS header, returns the true integration time
        for a file.
        """
        return head["T_INT"] / 1000.0

    def adjust_thisimage(self, thisimage):
        thisimage = thisimage.astype(float)
        return thisimage

    def read_data(self, raw_dir, new_dir):
        """
        Reads data.

        Inputs:
            :rawdir: (string) absolute path to directory containing raw data.
                        File path should end with '/'.
            :newdir: (string) absolute path to directory that will contain
                        4-quadrant data. File path should end with '/'.

        Outputs:
            None
        """

        def read_pharo(raw_file_name, new_file_name):
            """
            Read in the 4-quadrant data and flatten it.

            Inputs:
                :raw_file_name: (string) name of raw file.
                :new_file_name: (string) name of flattened 4-quadrant data.
            """
            im_cube = pyfits.getdata(raw_file_name)
            header = pyfits.getheader(raw_file_name)

            newfile = np.zeros((1024, 1024))
            newfile[0:512, 512:1024] = im_cube[0, :, :]  # Lower right
            newfile[0:512, 0:512] = im_cube[1, :, :]  # Lower left
            newfile[512:1024, 0:512] = im_cube[2, :, :]  # Upper left
            newfile[512:1024, 512:1024] = im_cube[3, :, :]  # Upper right

            hdu = pyfits.PrimaryHDU(newfile, header=header)
            hdu.writeto(new_file_name, overwrite=True, output_verify="ignore")

            return newfile

        def convert_night():
            """
            Convert the cubes to flat files for a whole night.
            """
            flist = glob.glob(raw_dir + "*.fits")

            if not os.path.isdir(new_dir):
                os.mkdir(new_dir)

            for fpath in flist:
                fname = fpath.split("/")[-1]
                newfname = new_dir + "s" + fname

                read_pharo(fpath, newfname)

        convert_night()



"""
Functions to work with flats.
"""
import astropy.io.fits as pyfits
import numpy as np
from tqdm import tqdm

from . import plotting as pl
from . import utils as u

CENTER = (750, 1100)  # row,col
MPIX = 600


def flat_driver(raw_dir, reddir, config, inst, plotting_yml=None):
    """Sets up and runs create_flats.

    Inputs:
        :raw_dir: (string) directory for the raw data
        :reddir: (string) directory for the reduced data
        :config: (pandas DataFrame) dataframe corresponding to config sheet for data.
        :inst: (Instrument object) instrument for which data is being reduced.
        :plot: (bool) determines whether or not intermediate plots should be produced.

    """
    if plotting_yml:
        pl.initialize_plotting(plotting_yml)

    _flats = config[config.Object == "flat"]
    filts = _flats.Filter.tolist()
    for filter_name in tqdm(
        filts, desc="Running flats", position=0, leave=True
    ):
        # literal_eval issues below
        flatlist = eval(
            _flats[_flats.Filter == filter_name].Filenums.values[0]
        )  # pylint: disable=eval-used
        itime = _flats[_flats.Filter == filter_name].ExpTime.values[0]
        darkfile = reddir + f"dark_{int(round(itime))}sec.fits"
        create_flats(
            raw_dir, reddir, flatlist, darkfile, inst, filter_name=filter_name
        )


def create_flats(
    raw_dir, reddir, flatlist, darkfile, inst, filter_name=None, test=False
):
    """Create a flat from a single list of flat files.

    Inputs:
        :raw_dir: (str) directory where the raw data is stored.
        :reddir: (str) directory where the reduced data is stored.
        :flatlist: (list) list of integers corresponding to flats.
        :inst: (inst object) instrument for which data is being reduced.
        :center: (tuple) two-element tuple dictating where the center of each image is.
        :npix: (int) number of pixels in image.
        :pl: (Bool) determines whether plots are produced.
        :filter: filter name given if head['FILT1NAM'] == 'Unknown'
    """
    nflats = len(flatlist)
    flatfiles = u.make_filelist(raw_dir, flatlist, inst)
    flat_array = u.read_imcube(flatfiles)
    flat_array = inst.adjust_array(flat_array, nflats)
    head = inst.head(flatfiles[0])
    filt = inst.filt(nflats, head, filter_name)

    if test:
        dark = 0.0
    else:
        dark = pyfits.getdata(darkfile, 0)
    for i in range(nflats):
        flat_array[i, :, :] = flat_array[i, :, :] - dark
        flat_array[i, :, :] = flat_array[i, :, :] / np.median(
            flat_array[i, :, :]
        )

    final_flat = np.median(flat_array, axis=0)
    final_flat = final_flat / np.median(final_flat)

    pl.plot_array(
        "intermediate", flat_array, -2.0, 2.0, reddir, f"flat_cube_{filt}.png"
    )

    # CDD update
    #  head.update('DATAFILE', str(flatlist)) #add all file names
    head.set("DATAFILE", str(flatlist))  # add all file names
    # end CDD update

    hdu = pyfits.PrimaryHDU(final_flat, header=head)
    hdu.writeto(
        reddir + f"flat_{filt}.fits", overwrite=True, output_verify="ignore"
    )

    return final_flat



# AS
# Created: 7/17/19
# Updated: 2/16/20
# Updated: 6/16/20

"""
Capability to check whether logsheet format is conducive to creating a config.
"""

import numpy as np
import pandas as pd

from . import add_dark_exp as ad


def check_logsheet(inst, log_name, tab=None, add_dark_times=False):
    """Checks for common typos/type errors in the logsheet. Should be
    run if an Excel worksheet is sent.

    Inputs:
        :inst: (Instrument object) instrument for which data is being reduced.
        :log_name: (string) path of the logsheet.
        :tab: (string) tab of interest,
        :add_dark_times: (bool) if true, runs the script within add_dark_exp to add the data from the
                        automated dark script to the log sheet.
    Outputs:
        :failed: (int) number of failed tests.
    """

    def check_tab(inst, add_dark_times, tab=None):
        """
        Checks for typos for one day in the observing run,
        assuming each day corresponds to a new sheet.

        Inputs:
            :inst: (Instrument object) instrument for which data is being reduced.
            :add_dark_times: (bool) if true, runs the script within add_dark_exp to add the data from the
                            automated dark script to the log sheet.
            :tab: (string) tab of interest. None for a CSV.


        """
        if add_dark_times:
            ad.add_dark_exp(inst, log, raw_dir, tab=None)
        frame_cols = log_frame.columns
        desired_cols = [
            "Object",
            "Start",
            "End",
            "Expose",
            "ExpTime",
            "Filter",
        ]
        missing = []
        failed = 0
        for col in desired_cols:
            if not np.isin(col, frame_cols):
                missing.append(col)
        if len(missing) != 0:
            print(f"Missing columns for {missing}.")
            failed += 1

        objects = log_frame["Object"].dropna().values
        exptimes = log_frame["ExpTime"].dropna().values
        if len(exptimes) != len(objects):
            print("Missing an exposure time.")
            failed += 1

        filters = log_frame["Filter"].dropna().values
        if len(filters) != len(log_frame[log_frame["Object"] != "dark"]):
            print("Missing a filter.")
            failed += 1

        starts = log_frame["Start"].dropna().values
        if len(starts) != len(objects):
            print("Missing a start exposure.")
            failed += 1

        ends = log_frame["End"].dropna().values
        if len(ends) != len(objects):
            print("Missing an end exposure.")
            failed += 1

        coadds = log_frame["Coadds"].dropna().values
        if len(coadds) != len(objects):
            print("Missing a coadd.")
            failed += 1

        if not np.all(exptimes > 0):
            print("There are negative or 0 exposure times.")
            failed += 1
        try:
            inter = ends - starts
            if not np.all(inter >= 0):
                print("Check the start and end exposures.")
                failed += 1
        except ValueError:
            print("Check the start and end exposures.")
            failed += 1

        exposes = log_frame["Expose"].dropna().values
        try:
            if not np.all(exposes == inter + 1):
                print(
                    "Incorrect number of exposures for start and end exposure."
                )
                failed += 1
        except UnboundLocalError:
            print("Incorrect number of exposures for start and end exposure.")
            failed += 1
        print(f"{9-failed}/9 tests passed.")
        return failed

    failed = 0
    if log_name[-3:] == "csv":
        log_frame = pd.read_csv(log_name)
        failed += check_tab(inst, add_dark_times=add_dark_times)
    elif log_name[-4:] == "xlsx" or log_name[-3:] == "xls":
        log = pd.ExcelFile(log_name)
        if not tab:
            for sheet in log.sheet_names:
                log_frame = pd.read_excel(log, sheet)
                failed += check_tab(
                    inst, add_dark_times=add_dark_times, tab=sheet
                )
        else:
            log_frame = pd.read_excel(log, tab)
            failed += check_tab(inst, add_dark_times=add_dark_times, tab=tab)
    return failed
